#!/usr/bin/env python3
"""
This example demonstrates how to fit a complex experimental setup using BornAgain.
It is based on real data published in  https://doi.org/10.1002/advs.201700856
by A. Glavic et al.
In this example we utilize the scalar reflectometry engine to fit polarized
data without spin-flip for performance reasons.
"""

import bornagain as ba, numpy as np, os, matplotlib.pyplot as plt, scipy
from bornagain import nm, sample_tools as st


datadir = os.getenv('BA_DATA_DIR', '')
if not datadir:
    raise Exception("Environment variable BA_DATA_DIR not set")

####################################################################
#  Sample and simulation model
####################################################################

def get_sample(P, sign, T):

    if T < 200:
        ms150 = P["ms150"]
    else:
        ms150 = 1

    air_color = (0.90, 0.93, 0.97)
    air_mat = ba.SLDMaterial("Air", air_color, 0, 0)
    pyox_color = (0.62, 0.68, 0.72)
    pyox_mat = ba.SLDMaterial("PyOx", pyox_color,
                               (P["sld_PyOx_real"] + \
                                 sign * ms150 * P["msld_PyOx"] )* 1e-6,
                               P["sld_PyOx_imag"] * 1e-6)
    py2_color = (0.58, 0.40, 0.74)
    py2_mat = ba.SLDMaterial("Py2", py2_color,
                               ( P["sld_Py2_real"] + \
                                 sign * ms150 * P["msld_Py2"] ) * 1e-6,
                               P["sld_Py2_imag"] * 1e-6)
    py1_color = (0.05, 0.62, 0.55)
    py1_mat = ba.SLDMaterial("Py1", py1_color,
                               ( P["sld_Py1_real"] + \
                                 sign * ms150 * P["msld_Py1"] ) * 1e-6,
                               P["sld_Py1_imag"] * 1e-6)
    sio2_color = (0.25, 0.74, 0.42)
    sio2_mat = ba.SLDMaterial("SiO2", sio2_color, P["sld_SiO2_real"]*1e-6,
                                     P["sld_SiO2_imag"]*1e-6)
    si_color = (0.28, 0.57, 0.82)
    si_mat = ba.SLDMaterial("Substrate", si_color, P["sld_Si_real"]*1e-6,
                                   P["sld_Si_imag"]*1e-6)

    transient_model = ba.ErfTransient()

    rPyOx_autocorr = ba.SelfAffineFractalModel(P["rPyOx"]*nm, 0.7, 25*nm)
    rPy2_autocorr = ba.SelfAffineFractalModel(P["rPy2"]*nm, 0.7, 25*nm)
    rPy1_autocorr = ba.SelfAffineFractalModel(P["rPy1"]*nm, 0.7, 25*nm)
    rSiO2_autocorr = ba.SelfAffineFractalModel(P["rSiO2"]*nm, 0.7, 25*nm)
    rSi_autocorr = ba.SelfAffineFractalModel(P["rSi"]*nm, 0.7, 25*nm)

    rPyOx = ba.Roughness(rPyOx_autocorr, transient_model)
    rPy2 = ba.Roughness(rPy2_autocorr, transient_model)
    rPy1 = ba.Roughness(rPy1_autocorr, transient_model)
    rSiO2 = ba.Roughness(rSiO2_autocorr, transient_model)
    rSi = ba.Roughness(rSi_autocorr, transient_model)

    l_Air = ba.Layer(air_mat)
    l_PyOx = ba.Layer(pyox_mat, P["t_PyOx"]*nm, rPyOx)
    l_Py2 = ba.Layer(py2_mat, P["t_Py2"]*nm, rPy2)
    l_Py1 = ba.Layer(py1_mat, P["t_Py1"]*nm, rPy1)
    l_SiO2 = ba.Layer(sio2_mat, P["t_SiO2"]*nm, rSiO2)
    l_Si = ba.Layer(si_mat, rSi)

    sample = ba.Sample()

    sample.addLayer(l_Air)
    sample.addLayer(l_PyOx)
    sample.addLayer(l_Py2)
    sample.addLayer(l_Py1)
    sample.addLayer(l_SiO2)
    sample.addLayer(l_Si)

    return sample


def run_simulation(qaxis, P, *, sign, T):

    qdistr = ba.DistributionGaussian(0., 1., 25, 3.)

    dq = P["dq"]*qaxis
    scan = ba.QzScan(qaxis)
    scan.setVectorResolution(qdistr, dq)
    scan.setIntensity(P["intensity"])

    sample = get_sample(P, sign, T)

    simulation = ba.SpecularSimulation(scan, sample)
    simulation.setBackground(ba.ConstantBackground(5e-7))

    return simulation.simulate().intensities()

####################################################################
#  Experimental data
####################################################################

def load_data(fname, qmin, qmax):
    fpath = os.path.join(datadir, fname)
    flags = ba.ImportSettings1D("q (1/angstrom)", "#", "", 1, 3, 4, 5)
    data = ba.readData1D(fpath, ba.csv1D, flags)
    data = data.normalizedToMax()
    return data.crop(qmin, qmax)

####################################################################
#  Plotting
####################################################################

def plot(q, rs, data, shifts, labels):
    """
    Plot the simulated result together with the experimental data.
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)

    for r, exp, shift, l in zip(rs, data, shifts, labels):

        ax.errorbar(exp.xCenters(),
                    exp.intensities() / shift,
                    yerr=exp.errors() / shift,
                    fmt='.',
                    markersize=0.75,
                    linewidth=0.5)

        ax.plot(q, r/shift, label=l)

    ax.set_yscale('log')
    plt.legend()

    plt.xlabel(r"$q\; $(nm$^{-1}$)")
    plt.ylabel("$R$")
    <%- if test_mode or figure_mode -%>
    <%- else -%>
    plt.tight_layout()
    <%- end -%>


def plot_sld_profile(P):

    z_300p, sld_300p = st.materialProfile(get_sample(P, +1, 300))
    z_300m, sld_300m = st.materialProfile(get_sample(P, -1, 300))
    z_150p, sld_150p = st.materialProfile(get_sample(P, +1, 150))
    z_150m, sld_150m = st.materialProfile(get_sample(P, -1, 150))

    plt.figure()
    plt.plot(z_300p, np.real(sld_300p)*1e6, label=r"300K $+$")
    plt.plot(z_300m, np.real(sld_300m)*1e6, label=r"300K $-$")
    plt.plot(z_150p, np.real(sld_150p)*1e6, label=r"150K $+$")
    plt.plot(z_150m, np.real(sld_150m)*1e6, label=r"150K $-$")

    plt.xlabel(r"$z\;$(Å)")
    plt.ylabel(r"$\delta(z) \cdot 10^6$")

    plt.legend()
    <%- if test_mode or figure_mode -%>
    <%- else -%>
    plt.tight_layout()
    <%- end -%>

####################################################################
#  Main
####################################################################

if __name__ == '__main__':

    # Parameters and bounds.

    # We start with rather good values so that the example takes not too much time
    startPnB = {
        "intensity": (0.5, 0.4, 0.6),
        "t_PyOx": (7.7, 6.0, 10.0),
        "t_Py2": (5.6, 4.6, 6.6),
        "t_Py1": (5.6, 4.6, 6.6),
        "t_SiO2": (2.2, 1.5, 2.9),
    }

    # For fixed parameters, bounds are ignored. We leave them here just
    # to facilitate moving entries between startPnB and fixedPnB.
    fixedPnB = {
        "sld_PyOx_imag": (0, 0, 0),
        "sld_Py2_imag": (0, 0, 0),
        "sld_Py1_imag": (0, 0, 0),
        "sld_SiO2_imag": (0, 0, 0),
        "sld_Si_imag": (0, 0, 0),
        "sld_SiO2_real": (3.47, 3, 4),
        "sld_Si_real": (2.0704, 2, 3),
        "dq": (0.018, 0, 0.1),
    # Start by moving the following back to startPnB:
        "sld_PyOx_real": (1.995, 1.92, 2.07),
        "sld_Py2_real": (5, 4.7, 5.3),
        "sld_Py1_real": (4.62, 4.32, 4.92),
        "rPyOx": (2.7, 1.5, 3.5),
        "rPy2": (1.2, .2, 2.0),
        "rPy1": (1.2, .2, 2.0),
        "rSiO2": (1.5, .5, 2.5),
        "rSi": (1.5, .5, 2.5),
        "msld_PyOx": (0.25, 0, 1),
        "msld_Py2": (0.63, 0, 1),
        "msld_Py1": (0.64, 0, 1),
        "ms150": (1.05, 1.0, 1.1),
    }

    fixedP = {d: v[0] for d, v in fixedPnB.items()}
    P = {d: v[0] for d, v in startPnB.items()} | fixedP
    bounds = [(par[1], par[2]) for par in startPnB.values()]
    freeParNames = [name for name in startPnB.keys()]

    # Restrict the q range for fitting and plotting
    qmin = 0.08
    qmax = 1.4

    data = [
        load_data("specular/honeycomb300p.dat", qmin, qmax),
        load_data("specular/honeycomb300m.dat", qmin, qmax),
        load_data("specular/honeycomb150p.dat", qmin, qmax),
        load_data("specular/honeycomb150m.dat", qmin, qmax)]

    simFunctions = [
        lambda q, P: run_simulation(q, P, sign=+1, T=300),
        lambda q, P: run_simulation(q, P, sign=-1, T=300),
        lambda q, P: run_simulation(q, P, sign=+1, T=150),
        lambda q, P: run_simulation(q, P, sign=-1, T=150)]

    qzs = np.linspace(qmin, qmax, 1500) # x-axis for plot R vs q

    # Plot data with initial model

    simResults = [ f(qzs, P) for f in simFunctions ]
    plot(qzs, simResults, data, [1, 1, 10, 10],
         ["300K $+$", "300K $-$", "150K $+$", "150K $-$"])
    plot_sld_profile(P)

    # Fit

    qaxes = [d.xCenters() for d in data]
    rdata = [d.intensities() for d in data]

    def par_dict(*args):
        return {name: value for name, value in zip(freeParNames, *args)} | fixedP

    def objective_function(*args):
        """
        Returns fit objective, i.e. sum of weighted squared relative differences.
        """
        fullP = par_dict(*args)
        result = 0
        for q, r, sim_fct in zip(qaxes, rdata, simFunctions):
            t = sim_fct(q, fullP)
            reldiff = (r - t) / (r + t)
            result += np.sum(reldiff**2/len(t))
        return result

    result = scipy.optimize.differential_evolution(
        objective_function,
        bounds,
        <%- if test_mode or figure_mode -%>
        maxiter=2,
        popsize=2,
        <%- else -%>
        maxiter=5, # for a serious DE fit, choose 500
        popsize=3, # for a serious DE fit, choose 10
        <%- end -%>
        tol=1e-2,
        mutation=(0.5, 1.5),
        seed=0,
        disp=True,
        <%- if test_mode or figure_mode -%>
        polish=False
        <%- else -%>
        polish=True
        <%- end -%>
    )

    print(f"Final chi2: {result.fun}")
    print("Fit Result:")
    for name, value in zip(freeParNames, result.x):
        print(f'   {name} = {value}')

    # Plot data with fit result

    P = par_dict(result.x)
    <%- unless test_mode || figure_mode -%>
    ba.showSample3D(get_sample(P, 1, 300), sample_size=120*nm, seed=0)
    <%- end -%>

    sim_results = [ f(qzs, P) for f in simFunctions ]
    plot(qzs, simResults, data, [1, 1, 10, 10],
         ["300K $+$", "300K $-$", "150K $+$", "150K $-$"])
    plot_sld_profile(P)

    <%- if test_mode or figure_mode -%>
    <%- else -%>
    plt.show()
    <%- end -%>



#!/usr/bin/env python3
"""
Dense cylinders, approximated as ideal gas or as hard disks.
Horizontal cut through GISAS image.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, nm2


def get_sample(approx):
    # materials
    vacuum = ba.Vacuum()
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-6, 2e-8)
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 3e-5, 2e-8)

    # particle
    ff = ba.Cylinder(5*nm, 1*nm)
    particle = ba.Particle(particle_mat, ff)

    # layers
    toplayer = ba.Layer(vacuum)
    density = 0.008/nm2
    if approx == "Dilute":
        toplayer.deposit2D(ba.Dilute2D(density, particle))
    else:
        toplayer.deposit2D(ba.Dense2D(density, particle))
    substrate = ba.Layer(substrate_mat)

    sample = ba.Sample()
    sample.addLayer(toplayer)
    sample.addLayer(substrate)
    return sample


def simulate(sample, title):
    beam = ba.Beam(1e9, 0.03*nm, 0.2*deg)
    n = 41
    detector = ba.SphericalDetector(n, -1*deg, 1*deg, 1, 0.78*deg, 0.82*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    field = simulation.simulate()
    field.setTitle(title)
    return field.flat()


if __name__ == '__main__':
    results = [
        simulate(get_sample("Dilute"), "ideal gas"),
        simulate(get_sample("Dense"), "hard disks"),
    ]
    from bornagain import ba_check
    ba_check.persistence_test(results)



#!/usr/bin/env python3
"""
Square lattice of half spheres on substrate.
Test to investigate intensity changes with different numbers of slices.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm
import matplotlib.pyplot as plt


def get_sample(n_slices=1):
    # Materials (zero absorption)
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 3e-05, 0)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-06, 0)
    vacuum = ba.Vacuum()

    # Particles
    ff = ba.SphericalSegment(5*nm, 0, 5*nm)
    particle = ba.Particle(particle_mat, ff)

    # 2D lattices
    lattice = ba.BasicLattice2D(10*nm, 10*nm, 90*deg, 0)

    # Interference functions
    layout = ba.Crystal2D(particle, lattice)
    profile = ba.Profile2DCauchy(100*nm, 100*nm, 0)
    layout.setDecayFunction(profile)

    # Layers
    layer_1 = ba.Layer(vacuum)
    if n_slices > 1:
        layer_1.setNumberOfSlices(n_slices)
    layer_1.deposit2D(layout)
    layer_2 = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def simulate_horizontal_cut(n_slices=1):
    """Simulate with Nx1 detector for horizontal cut at alpha_f=0.92 deg."""
    sample = get_sample(n_slices)
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 200
    # Single bin centered at alpha_f=0.92 deg
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, 1, 0.91*deg, 0.93*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation.simulate()


def simulate_vertical_cut(n_slices=1):
    """Simulate with 1xN detector for vertical cut at phi_f=0 deg."""
    sample = get_sample(n_slices)
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 200
    # Single bin centered at phi_f=0 deg
    detector = ba.SphericalDetector(1, -0.01*deg, 0.01*deg, n, 0, 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation.simulate()


def simulate_depthprobe(n_slices=1):
    """Simulate intensity vs depth."""
    sample = get_sample(n_slices)
    n = 100
    alpha_f = 0.03*deg
    scan = ba.AlphaScan([alpha_f])
    scan.setWavelength(0.1*nm)
    z_axis = ba.EquiDivision("z (nm)", n, -20*nm, 120*nm)
    simulation = ba.DepthprobeSimulation(scan, sample, z_axis)
    return simulation.simulate(), alpha_f/deg


# Color cycle for different N values
colors = ['C0', 'C1', 'C2', 'C3']


if __name__ == '__main__':
    # Slice counts to compare
    slice_counts = [1, 2, 5, 20]

    # Create figure with 1 row, 3 columns
    fig, axes = plt.subplots(1, 3, figsize=(15, 5))

    # Horizontal cuts at alpha_f=0.92 deg
    ax = axes[0]
    for i, n in enumerate(slice_counts):
        result = simulate_horizontal_cut(n_slices=n)
        ax.semilogy(result.xCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title('Horizontal cut at alpha_f=0.92 deg')
    ax.set_xlabel('phi_f (deg)')
    ax.set_ylabel('Intensity')
    ax.legend()

    # Vertical cuts at phi_f=0 deg
    ax = axes[1]
    for i, n in enumerate(slice_counts):
        result = simulate_vertical_cut(n_slices=n)
        ax.semilogy(result.yCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title('Vertical cut at phi_f=0 deg')
    ax.set_xlabel('alpha_f (deg)')
    ax.set_ylabel('Intensity')
    ax.legend()

    # Intensity vs depth
    ax = axes[2]
    for i, n in enumerate(slice_counts):
        result, alpha_f = simulate_depthprobe(n_slices=n)
        ax.semilogy(result.yCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title(f'Intensity vs depth at alpha_f={alpha_f:.4f} deg')
    ax.set_xlabel('z (nm)')
    ax.set_ylabel('Intensity')
    ax.legend()

    plt.tight_layout()
    plotargs = bp.parse_commandline()
    bp.export(**plotargs)



#!/usr/bin/env python3
"""
Reorder member functions in a .cpp file to match the order in corresponding .h file.
Uses libclang for AST-based parsing instead of regex.

⚠️  NOT APPLICABLE TO QT-BASED CODE ⚠️

This script uses libclang to parse C++ files via AST (Abstract Syntax Tree) analysis,
which is theoretically superior to regex-based parsing as it properly handles:
- Preprocessor directives
- Complex C++ syntax (templates, nested classes, etc.)
- Comments and string literals
- Function overloads and const variants

However, it CANNOT parse BornAgain code due to Qt dependencies:

WHY IT FAILS:
1. Qt Headers: BornAgain headers include Qt types (QString, QWidget, etc.)
2. Missing Type Resolution: libclang cannot resolve Qt types without full Qt headers
3. Preprocessing Complexity: Qt uses heavy macro preprocessing (Q_OBJECT, signals, slots)
4. moc Dependencies: Qt's Meta-Object Compiler generates additional code
5. Library Mismatch: Bundled libclang may not properly locate system Qt headers

ATTEMPTS MADE:
- Added Qt include paths: Failed, Qt headers have circular dependencies
- Used compilation database: Failed, libclang raises TranslationUnitLoadError
- Stubbed out includes: Failed, types still unresolved
- Used PARSE_INCOMPLETE option: Still fails before building AST

POTENTIAL SOLUTIONS (not implemented):
- Preprocess files first: clang++ -E to expand all includes/macros
- Use LibTooling directly: C++ tool with better compilation database integration
- Mock Qt headers: Create stub declarations for all Qt types used
- Different Python bindings: Try alternatives to pip's libclang package

For BornAgain, use reorder-functions-regex.py instead, which handles most cases
correctly using pattern matching (though it has edge cases with preprocessor
directives and complex initialization).

Usage:
    reorder-functions-ast.py <file_stem>
"""

import sys
from typing import List, Tuple, Dict, Optional
from dataclasses import dataclass
import clang.cindex
from clang.cindex import CursorKind, AccessSpecifier

@dataclass
class FunctionDecl:
    """Represents a function declaration from a header file."""
    class_name: str
    function_name: str
    signature: str  # Canonical signature for matching
    line_number: int
    is_const: bool
    is_virtual: bool
    is_pure_virtual: bool
    is_static: bool

@dataclass
class FunctionDef:
    """Represents a function definition from a cpp file."""
    class_name: str
    function_name: str
    signature: str  # Canonical signature for matching
    start_offset: int
    end_offset: int
    is_const: bool
    is_static: bool
    full_text: str

def init_libclang():
    """Initialize libclang with system includes."""
    # Try to find libclang automatically, or use libclang-19
    import os
    import glob

    # Check if libclang package provides it
    try:
        from clang.cindex import conf
        # libclang package comes with its own library
        return
    except:
        pass

    # Search for system libclang
    search_paths = [
        '/usr/lib/x86_64-linux-gnu/libclang-*.so.1',
        '/usr/lib/llvm-*/lib/libclang-*.so.1',
    ]

    for pattern in search_paths:
        matches = glob.glob(pattern)
        if matches:
            # Use the first match (typically highest version)
            clang.cindex.Config.set_library_file(sorted(matches)[-1])
            return

    # If nothing found, let libclang try its defaults
    pass

def get_canonical_signature(cursor) -> str:
    """
    Get a canonical signature string for a function.
    This is used for matching declarations to definitions.
    """
    # Get parameter types
    param_types = []
    for arg in cursor.get_arguments():
        param_types.append(arg.type.spelling)

    # Include const qualifier
    const_str = " const" if cursor.is_const_method() else ""

    return f"{cursor.spelling}({', '.join(param_types)}){const_str}"

def is_constructor_or_destructor(cursor, class_name: str) -> bool:
    """Check if cursor is a constructor or destructor."""
    name = cursor.spelling
    return name == class_name or name == f"~{class_name}"

def get_compile_args() -> List[str]:
    """Get compilation arguments by reading from compilation database or using defaults."""
    import json
    import os

    # Try to read from compilation database
    comp_db_path = '/G/sw/ba/build/compile_commands.json'
    if os.path.exists(comp_db_path):
        try:
            with open(comp_db_path) as f:
                commands = json.load(f)
            if commands:
                # Extract include paths and defines from first command
                args = []
                for cmd_entry in commands:
                    cmd = cmd_entry.get('command', '')
                    parts = cmd.split()
                    for i, part in enumerate(parts):
                        if part.startswith('-I') or part.startswith('-D') or part == '-std=c++20':
                            if part in ['-I', '-D'] and i + 1 < len(parts):
                                args.append(part + parts[i+1])
                            else:
                                args.append(part)
                if args:
                    return list(set(args))  # Remove duplicates
        except:
            pass

    # Default args
    return [
        '-std=c++20',
        '-I/G/sw/ba',
        '-I/G/sw/ba/build/inc',
        '-I/G/sw/ba/3rdparty/Core',
        '-I/usr/include/x86_64-linux-gnu/qt6',
        '-I/usr/include/x86_64-linux-gnu/qt6/QtCore',
        '-I/usr/include/x86_64-linux-gnu/qt6/QtWidgets',
        '-I/usr/include/x86_64-linux-gnu/qt6/QtGui',
        '-DQT_CORE_LIB',
        '-DBORNAGAIN_PYTHON',
        '-w',  # Suppress warnings
    ]

def extract_function_declarations(header_file: str) -> List[FunctionDecl]:
    """Extract member function declarations from header file using AST."""
    index = clang.cindex.Index.create()

    args = get_compile_args()

    # Parse with options to skip function bodies and continue on errors
    options = (
        clang.cindex.TranslationUnit.PARSE_SKIP_FUNCTION_BODIES |
        clang.cindex.TranslationUnit.PARSE_INCOMPLETE
    )

    try:
        tu = index.parse(header_file, args=args, options=options)
    except clang.cindex.TranslationUnitLoadError:
        # Try with even more lenient options
        print(f"Warning: Initial parse failed, trying with minimal options...")
        args = ['-std=c++20', '-w']
        tu = index.parse(header_file, args=args, options=options)

    if not tu:
        raise RuntimeError(f"Failed to parse {header_file}")

    # Check for fatal errors
    fatal_errors = [d for d in tu.diagnostics
                   if d.severity >= clang.cindex.Diagnostic.Error]
    if fatal_errors:
        print(f"Warning: {len(fatal_errors)} errors while parsing {header_file}:")
        for diag in fatal_errors[:5]:  # Show first 5
            print(f"  {diag}")
        # Continue anyway - we might still extract useful information

    declarations = []

    def visit_class(cursor, class_name):
        """Visit members of a class."""
        for child in cursor.get_children():
            # Only look at member functions
            if child.kind != CursorKind.CXX_METHOD:
                continue

            # Skip if defined inline in header
            if child.is_definition():
                continue

            # Skip constructors and destructors
            if is_constructor_or_destructor(child, class_name):
                continue

            # Skip operators
            if child.spelling.startswith('operator'):
                continue

            # Check if pure virtual
            is_pure = child.is_pure_virtual_method()

            # Skip pure virtual functions
            if is_pure:
                continue

            # Get location
            location = child.location
            if not location.file or location.file.name != header_file:
                continue

            decl = FunctionDecl(
                class_name=class_name,
                function_name=child.spelling,
                signature=get_canonical_signature(child),
                line_number=location.line,
                is_const=child.is_const_method(),
                is_virtual=child.is_virtual_method(),
                is_pure_virtual=is_pure,
                is_static=child.is_static_method()
            )
            declarations.append(decl)

    def visit_node(cursor):
        """Recursively visit AST nodes."""
        if cursor.kind == CursorKind.CLASS_DECL or cursor.kind == CursorKind.STRUCT_DECL:
            # Found a class/struct definition
            class_name = cursor.spelling
            if class_name:  # Skip anonymous classes
                visit_class(cursor, class_name)

        # Recurse into children
        for child in cursor.get_children():
            # Only visit declarations in this file
            if child.location.file and child.location.file.name == header_file:
                visit_node(child)

    visit_node(tu.cursor)

    return declarations

def extract_function_definitions(cpp_file: str) -> List[FunctionDef]:
    """Extract member function definitions from cpp file using AST."""
    index = clang.cindex.Index.create()

    args = get_compile_args()

    # Parse with options to continue on errors
    options = clang.cindex.TranslationUnit.PARSE_INCOMPLETE

    try:
        tu = index.parse(cpp_file, args=args, options=options)
    except clang.cindex.TranslationUnitLoadError:
        # Try with minimal options
        print(f"Warning: Initial parse failed, trying with minimal options...")
        args = ['-std=c++20', '-w']
        tu = index.parse(cpp_file, args=args, options=options)

    if not tu:
        raise RuntimeError(f"Failed to parse {cpp_file}")

    # Check for fatal errors
    fatal_errors = [d for d in tu.diagnostics
                   if d.severity >= clang.cindex.Diagnostic.Error]
    if fatal_errors:
        print(f"Warning: {len(fatal_errors)} errors while parsing {cpp_file}:")
        for diag in fatal_errors[:5]:  # Show first 5
            print(f"  {diag}")
        # Continue anyway

    # Read the source file for text extraction
    with open(cpp_file, 'r') as f:
        source_text = f.read()

    definitions = []

    def visit_node(cursor):
        """Recursively visit AST nodes."""
        # Look for CXX_METHOD that is a definition (not declaration)
        if cursor.kind == CursorKind.CXX_METHOD and cursor.is_definition():
            # Check if this is in the cpp file (not in included headers)
            location = cursor.location
            if not location.file or location.file.name != cpp_file:
                for child in cursor.get_children():
                    visit_node(child)
                return

            # Get the semantic parent (the class)
            semantic_parent = cursor.semantic_parent
            if semantic_parent.kind not in (CursorKind.CLASS_DECL, CursorKind.STRUCT_DECL):
                for child in cursor.get_children():
                    visit_node(child)
                return

            class_name = semantic_parent.spelling

            # Skip constructors and destructors
            if is_constructor_or_destructor(cursor, class_name):
                for child in cursor.get_children():
                    visit_node(child)
                return

            # Skip operators
            if cursor.spelling.startswith('operator'):
                for child in cursor.get_children():
                    visit_node(child)
                return

            # Get extent (source range)
            extent = cursor.extent
            start_offset = extent.start.offset
            end_offset = extent.end.offset

            # Extract the full text
            full_text = source_text[start_offset:end_offset]

            # Make sure we capture complete lines
            # Walk back to start of line
            while start_offset > 0 and source_text[start_offset - 1] not in '\n':
                start_offset -= 1

            # Walk forward to end of line (including newlines after closing brace)
            while end_offset < len(source_text) and source_text[end_offset] in '\n':
                end_offset += 1

            full_text = source_text[start_offset:end_offset]

            definition = FunctionDef(
                class_name=class_name,
                function_name=cursor.spelling,
                signature=get_canonical_signature(cursor),
                start_offset=start_offset,
                end_offset=end_offset,
                is_const=cursor.is_const_method(),
                is_static=cursor.is_static_method(),
                full_text=full_text
            )
            definitions.append(definition)

        # Recurse into children
        for child in cursor.get_children():
            visit_node(child)

    visit_node(tu.cursor)

    return definitions

def match_functions(declarations: List[FunctionDecl],
                   definitions: List[FunctionDef]) -> List[Tuple[FunctionDecl, FunctionDef]]:
    """Match declarations from header to definitions in cpp."""
    matches = []

    # Build a lookup dict for definitions
    def_dict = {}
    for defn in definitions:
        key = (defn.class_name, defn.signature)
        if key in def_dict:
            # Handle overloads - store as list
            if not isinstance(def_dict[key], list):
                def_dict[key] = [def_dict[key]]
            def_dict[key].append(defn)
        else:
            def_dict[key] = defn

    # Match each declaration
    for decl in declarations:
        key = (decl.class_name, decl.signature)
        if key in def_dict:
            defn = def_dict[key]
            if isinstance(defn, list):
                # Multiple definitions - take first one
                defn = defn[0]
            matches.append((decl, defn))

    return matches

def reorder_cpp_file(header_file: str, cpp_file: str) -> bool:
    """Reorder functions in cpp_file to match order in header_file."""
    init_libclang()

    print(f"Parsing {header_file}...")
    declarations = extract_function_declarations(header_file)

    if not declarations:
        print(f"No member functions found in {header_file}")
        return False

    print(f"Found {len(declarations)} declarations in header")

    print(f"Parsing {cpp_file}...")
    definitions = extract_function_definitions(cpp_file)

    if not definitions:
        print(f"No member function implementations found in {cpp_file}")
        return False

    print(f"Found {len(definitions)} definitions in cpp")

    # Match declarations to definitions
    matches = match_functions(declarations, definitions)

    if not matches:
        print(f"No matching functions between {header_file} and {cpp_file}")
        return False

    print(f"Matched {len(matches)} functions")

    # Check if already in order
    current_order = sorted(matches, key=lambda m: m[1].start_offset)
    if matches == current_order:
        print(f"Functions already in correct order in {cpp_file}")
        return False

    # Read the source file
    with open(cpp_file, 'r') as f:
        source_text = f.read()

    # Find the span of all matched functions
    all_defs = [m[1] for m in matches]
    first_offset = min(d.start_offset for d in all_defs)
    last_offset = max(d.end_offset for d in all_defs)

    # Build reordered content
    result = source_text[:first_offset]

    for decl, defn in matches:
        result += defn.full_text
        if not defn.full_text.endswith('\n'):
            result += '\n'

    result += source_text[last_offset:]

    # Write back
    with open(cpp_file, 'w') as f:
        f.write(result)

    print(f"Reordered {len(matches)} functions in {cpp_file}")
    return True

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage: reorder-functions-ast.py <file_stem>")
        sys.exit(1)

    stem = sys.argv[1]
    try:
        changed = reorder_cpp_file(stem + '.h', stem + '.cpp')
        sys.exit(0 if changed else 1)
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        import traceback
        traceback.print_exc()
        sys.exit(2)



#!/usr/bin/env python3
"""
Cylindrical mesocrystal with hexagonal lattice and two-material basis.
Test to investigate intensity changes with different numbers of slices.
"""
import math
import numpy as np
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, nm2, R3
import matplotlib.pyplot as plt


def get_sample(n_slices=1):
    # Materials (zero absorption)
    particle1_color = (0.86, 0.24, 0.18)
    particle1_mat = ba.RefractiveMaterial("Particle1", particle1_color, 4e-05, 0)
    particle2_color = (0.25, 0.65, 0.35)
    particle2_mat = ba.RefractiveMaterial("Particle2", particle2_color, 8e-06, 0)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-06, 0)
    vacuum = ba.Vacuum()

    # Basis particles
    inner_ff = ba.FuzzySphere(4.7*nm, .4*nm)

    inner_particle1 = ba.Particle(particle1_mat, inner_ff)
    inner_particle2 = ba.Particle(particle2_mat, inner_ff)

    # 3D lattice
    lattice_a = 6.21
    lattice_c = 6.57
    sigma_a = 1.16
    position_variance = sigma_a/math.sqrt(3)

    lattice = ba.HexagonalLattice(lattice_a*2, lattice_c*2*2.3)

    bas_a = lattice.basisVectorA()
    bas_b = lattice.basisVectorB()
    bas_c = lattice.basisVectorC()

    # Basis composition
    position_0 = R3(0, 0, 0)
    position_1 = 1.0/3*(2*bas_a + bas_b + bas_c)
    basis = ba.Compound()
    basis.addComponent(inner_particle1, position_0)
    basis.addComponent(inner_particle2, position_1)

    # Crystal
    crystal = ba.Crystal(basis, lattice, position_variance)

    # Mesocrystal: shaped crystal
    outer_ff = ba.Cylinder(20*nm, 50*nm)
    outer_particle = ba.Mesocrystal(crystal, outer_ff)

    # Layers
    layer_1 = ba.Layer(vacuum)
    layer_1.deposit2D(ba.Dilute2D(0.0001/nm2, outer_particle))
    if n_slices > 1:
        layer_1.setNumberOfSlices(n_slices)
    layer_2 = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def simulate_horizontal_cut(n_slices=1, reciprocal=False):
    """Simulate with Nx1 detector for horizontal cut at alpha_f=0.92 deg."""
    sample = get_sample(n_slices)
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = <%= test_mode ? 11 : 200 %>
    # Single bin centered at alpha_f=0.92 deg
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, 1, 0.91*deg, 0.93*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    if reciprocal:
        simulation.options().setMesoReciprocalSum(True, 2.5)
    return simulation.simulate()


def simulate_vertical_cut(n_slices=1, reciprocal=False):
    """Simulate with 1xN detector for vertical cut at phi_f=0 deg."""
    sample = get_sample(n_slices)
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = <%= test_mode ? 11 : 200 %>
    # Single bin centered at phi_f=0 deg
    detector = ba.SphericalDetector(1, -0.01*deg, 0.01*deg, n, 0, 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    if reciprocal:
        simulation.options().setMesoReciprocalSum(True, 2.5)
    return simulation.simulate()


def simulate_depthprobe(n_slices=1, reciprocal=False):
    """Simulate intensity vs depth."""
    sample = get_sample(n_slices)
    n = <%= test_mode ? 11 : 100 %>
    alpha_f = 0.001*deg if reciprocal else 0.0064*deg
    scan = ba.AlphaScan([alpha_f])
    scan.setWavelength(0.1*nm)
    z_axis = ba.EquiDivision("z (nm)", n, -20*nm, 120*nm)
    simulation = ba.DepthprobeSimulation(scan, sample, z_axis)
    if reciprocal:
        simulation.options().setMesoReciprocalSum(True, 2.5)
    return simulation.simulate(), alpha_f/deg


def get_sld_profile(n_slices=1):
    """Get SLD profile from resampled sample."""
    sample = get_sample(n_slices)
    n = <%= test_mode ? 11 : 200 %>
    z_min, z_max = -20*nm, 60*nm
    z = np.linspace(z_min, z_max, n)
    sld = ba.materialProfileSLD(sample, n, z_min, z_max)
    sld_real = np.array([s.real for s in sld])
    return z/nm, sld_real


# Color cycle for different N values
colors = ['C0', 'C1', 'C2', 'C3']


if __name__ == '__main__':
    # Slice counts to compare
    slice_counts = [1, 2, 5, 20]

    # Create figure with 2 rows, 4 columns
    fig, axes = plt.subplots(2, 4, figsize=(18, 10))

    # Row 1: Real-space summation (default)
    ax = axes[0, 0]
    for i, n in enumerate(slice_counts):
        result = simulate_horizontal_cut(n_slices=n, reciprocal=False)
        ax.semilogy(result.xCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title('Real-space: Horizontal cut at alpha_f=0.92 deg')
    ax.set_xlabel('phi_f (deg)')
    ax.set_ylabel('Intensity')
    ax.legend()

    ax = axes[0, 1]
    for i, n in enumerate(slice_counts):
        result = simulate_vertical_cut(n_slices=n, reciprocal=False)
        ax.semilogy(result.yCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title('Real-space: Vertical cut at phi_f=0 deg')
    ax.set_xlabel('alpha_f (deg)')
    ax.set_ylabel('Intensity')
    ax.legend()

    ax = axes[0, 2]
    for i, n in enumerate(slice_counts):
        result, alpha_f = simulate_depthprobe(n_slices=n, reciprocal=False)
        ax.semilogy(result.yCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title(f'Real-space: Intensity vs depth at alpha_f={alpha_f:.4f} deg')
    ax.set_xlabel('z (nm)')
    ax.set_ylabel('Intensity')
    ax.legend()

    ax = axes[0, 3]
    for i, n in enumerate(slice_counts):
        z, sld = get_sld_profile(n_slices=n)
        ax.plot(z, sld, label=f'N={n}', color=colors[i])
    ax.set_title('Real-space: SLD profile')
    ax.set_xlabel('z (nm)')
    ax.set_ylabel('SLD (real part)')
    ax.legend()

    # Row 2: Reciprocal-space (Fourier) summation
    ax = axes[1, 0]
    for i, n in enumerate(slice_counts):
        result = simulate_horizontal_cut(n_slices=n, reciprocal=True)
        ax.semilogy(result.xCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title('Fourier: Horizontal cut at alpha_f=0.92 deg')
    ax.set_xlabel('phi_f (deg)')
    ax.set_ylabel('Intensity')
    ax.legend()

    ax = axes[1, 1]
    for i, n in enumerate(slice_counts):
        result = simulate_vertical_cut(n_slices=n, reciprocal=True)
        ax.semilogy(result.yCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title('Fourier: Vertical cut at phi_f=0 deg')
    ax.set_xlabel('alpha_f (deg)')
    ax.set_ylabel('Intensity')
    ax.legend()

    ax = axes[1, 2]
    for i, n in enumerate(slice_counts):
        result, alpha_f = simulate_depthprobe(n_slices=n, reciprocal=True)
        ax.semilogy(result.yCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title(f'Fourier: Intensity vs depth at alpha_f={alpha_f:.3f} deg')
    ax.set_xlabel('z (nm)')
    ax.set_ylabel('Intensity')
    ax.legend()

    ax = axes[1, 3]
    for i, n in enumerate(slice_counts):
        z, sld = get_sld_profile(n_slices=n)
        ax.plot(z, sld, label=f'N={n}', color=colors[i])
    ax.set_title('Fourier: SLD profile')
    ax.set_xlabel('z (nm)')
    ax.set_ylabel('SLD (real part)')
    ax.legend()

    plt.tight_layout()
<%- if test_mode -%>
    # Just run, no reference check
    pass
<%- elsif figure_mode -%>
    plotargs = bp.parse_commandline()
    bp.export(**plotargs)
<%- else -%>
    ba.showSample3D(get_sample(), sample_size=250*nm, seed=0)
    plt.show()
<%- end -%>



#!/usr/bin/env python3
"""
Soft spheres (FuzzySphere) in a finite layer.

Soft particles have diffuse boundaries modelled by a Debye-Waller factor.
Slicing is not supported for soft particles; material averaging is disabled.

The layer thickness is 8*R. The particle center is shifted up by 4*R (= thickness/2)
so that almost all of the particle sits inside the finite layer.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, nm2


def get_sample():
    R = 5 * nm
    thickness = 8 * R  # 40 nm

    # Materials
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-04, 2e-08)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-06, 2e-08)
    vacuum = ba.Vacuum()

    # Soft particle (FuzzySphere with Gaussian fuzziness)
    ff = ba.FuzzySphere(R, 0.5 * nm)
    particle = ba.Particle(particle_mat, ff)
    # Shift center to thickness/2 (deposit2D anchors the bottom, so compensate)
    particle.translate(0, 0, thickness / 2 - particle.height_below_center())

    # Layers: vacuum / finite film / substrate
    layer_vacuum = ba.Layer(vacuum)
    layer_film = ba.Layer(vacuum, thickness)
    layer_film.deposit2D(ba.Dilute2D(1e-3 / nm2, particle))
    layer_substrate = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_vacuum)
    sample.addLayer(layer_film)
    sample.addLayer(layer_substrate)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1 * nm, 0.2 * deg)
    n = 200
    detector = ba.SphericalDetector(n, -2 * deg, 2 * deg, n, 0, 3 * deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    # Disable material averaging (slicing not supported for soft particles)
    simulation.options().setUseAvgMaterials(False)
    return simulation

if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    plotargs = bp.parse_commandline()
    bp.plot_datafield(result, unit_aspect=1, **plotargs)
    bp.export(**plotargs)



#!/usr/bin/env python3
"""
GISAS of cylinders with a magnetic material at zero field.

A cylinder is assigned a magnetic material with a zero magnetic field
vector. The unpolarized simulation is expected to agree with the
non-magnetic case, verifying that zero-field magnetic materials reduce
to ordinary refraction.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, R3


def get_sample():
    vacuum = ba.Vacuum()
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-6, 2e-8)
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial(
        "MagParticle", particle_color, 6e-4, 2e-8, R3(0, 0, 0))

    ff = ba.Cylinder(5*nm, 5*nm)
    particle = ba.Particle(particle_mat, ff)

    vacuum_layer = ba.Layer(vacuum)
    vacuum_layer.deposit2D(ba.Dilute2D(0.001, particle))
    substrate_layer = ba.Layer(substrate_mat)

    sample = ba.Sample()
    sample.addLayer(vacuum_layer)
    sample.addLayer(substrate_layer)
    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 200
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, n, 0, 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    plotargs = bp.parse_commandline()
    bp.plot_datafield(result, unit_aspect=1, **plotargs)
    bp.export(**plotargs)



#!/usr/bin/env python3
"""
Plot |F(q)| vs q for sphere.

"""
import bornagain as ba
import cmath
import numpy as np
from bornagain import ba_plot as bp, deg, nm, C3

if __name__ == '__main__':
    R = 1
    ff = ba.Sphere(R)
    qmax = 32

    v = ff.volume()
    n = 1000
    x = [qmax*i/(n - 1) for i in range(n)]
    # Undo the bottom-to-center shift built into Sphere's formfactor
    y = [(ff.formfactorAtBottom(C3(q, 0, 0)) * cmath.exp(-1j*q*R)).real/v for q in x]
    ym = [-f for f in y]

    bp.plt.semilogy(x, y)
    bp.plt.semilogy(x, ym)

    bp.plt.xlim([0, qmax])
    bp.plt.ylim([5e-5, 2])

    label_fontsize = 18
    bp.plt.xlabel("$qR$", fontsize=label_fontsize)
    bp.plt.ylabel("$|F(q)|$", fontsize=label_fontsize)

    bp.plt.tight_layout()
    bp.plt.show()



#!/usr/bin/env python3
"""
Reflectivity of a multilayer with different wavelength distributions.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, std_samples, nm


lambda0 = 0.154

def get_sample():
    return std_samples.alternating_layers()


def simulate(sample, distr, title):
    n = 300
    scan = ba.AlphaScan(n, 0.5*deg, 1*deg)
    scan.setWavelength(lambda0)
    if distr:
        scan.setWavelengthDistribution(distr)

    result = ba.SpecularSimulation(scan, sample).simulate()
    result.setTitle(title)
    return result


if __name__ == '__main__':
    sample = get_sample()
    ba.showSample3D(sample, sample_size=120*nm, seed=0)
    results = [
        simulate(sample, ba.DistributionLorentz(
            lambda0, 0.01177 * lambda0), "Lorentz"),
        simulate(sample, ba.DistributionGaussian(
            lambda0, 0.01 * lambda0), "Gauss"),
        simulate(sample, ba.DistributionLogNormal(
            lambda0, 0.011), "LogNormal"),
    ]
    bp.plot_multicurve(results)
    bp.plt.show()



#!/usr/bin/env python3
"""
Off-specular scattering from a Ti/Pt resonator on Si with D2O supernatant.

Sample: Si | 3x(Ti(13nm) + Pt(32nm)) | Ti_top(10nm) | TiO2(3nm) | D2O
All interfaces share SelfAffineFractal roughness (sigma=2nm, Hurst=0.8,
corrLen=1e4nm) with TanhTransient and CommonDepthCrosscorrelation (400nm).
"""
import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg, nm


def get_sample():
    m_si_color = (0.30, 0.62, 0.86)
    m_si = ba.RefractiveMaterial("Si", m_si_color, 8.25218379931e-06, 0)
    m_ti_color = (0.05, 0.62, 0.55)
    m_ti = ba.RefractiveMaterial(
        "Ti", m_ti_color, -7.6593316363e-06, 3.81961616312e-09)
    m_tio2_color = (0.48, 0.32, 0.80)
    m_tio2 = ba.RefractiveMaterial(
        "TiO2", m_tio2_color, 1.04803530026e-05, 2.03233519385e-09)
    m_pt_color = (0.93, 0.72, 0.25)
    m_pt = ba.RefractiveMaterial(
        "Pt", m_pt_color, 2.52936993309e-05, 7.54553992473e-09)
    m_d2o_color = (0.90, 0.93, 0.97)
    m_d2o = ba.RefractiveMaterial(
        "D2O", m_d2o_color, 2.52897204573e-05, 4.5224432814e-13)

    autocorr = ba.SelfAffineFractalModel(2*nm, 0.8, 1e4*nm)
    transient = ba.TanhTransient()
    crosscorr = ba.CommonDepthCrosscorrelation(400*nm)
    roughness = ba.Roughness(autocorr, transient, crosscorr)

    ti_thickness = 13*nm
    l_ti = ba.Layer(m_ti, ti_thickness, roughness)
    l_pt = ba.Layer(m_pt, 32*nm, roughness)
    l_ti_top = ba.Layer(m_ti, 10*nm, roughness)
    l_tio2 = ba.Layer(m_tio2, 3*nm, roughness)

    sample = ba.Sample()
    sample.addLayer(ba.Layer(m_si))
    for _ in range(3):
        sample.addLayer(l_ti)
        sample.addLayer(l_pt)
    sample.addLayer(l_ti_top)
    sample.addLayer(l_tio2)
    sample.addLayer(ba.Layer(m_d2o, roughness))
    return sample


def get_simulation(sample):
    nscan = 20
    ndet = 25
    scan = ba.AlphaScan(nscan, 0.2*deg, 4*deg)
    scan.setWavelength(5*angstrom)
    scan.setIntensity(1e9)
    detector = ba.OffspecDetector(ndet, 0.1*deg, 5.1*deg, -0.1*deg, 0.1*deg)
    simulation = ba.OffspecSimulation(scan, sample, detector)
    simulation.options().setIncludeSpecular(True)
    return simulation


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    from bornagain import ba_check
    ba_check.persistence_test(result)



#!/usr/bin/env python3
"""
Realistic (imperfect) example for polarized reflectometry.
Sample contains a magnetic layer,
similar to Devishvili et al., Rev. Sci. Instrum. 84, 025112 (2013).
"""
import numpy
import bornagain as ba
from bornagain import nm, ba_plot as bp, deg, R3
from math import sin, cos

def get_sample():
    # Materials
    Bmag = 1.6e6
    Bangle = 0 * deg
    B = R3(Bmag*sin(Bangle), Bmag*cos(Bangle), 0)
    vacuum = ba.Vacuum()
    pd_color = (0.93, 0.72, 0.25)
    pd_mat = ba.SLDMaterial("Pd", pd_color, 4.0099e-6, 1.3019e-09)
    fe_color = (0.48, 0.32, 0.80)
    fe_mat = ba.SLDMaterial("Fe", fe_color, 8.0241e-06, 6.0448e-10, B)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.SLDMaterial("MgO", substrate_color, 5.9803e-06, 9.3996e-12)

    autocorr = ba.SelfAffineFractalModel(2*nm, 0.7, 25*nm)
    transient = ba.TanhTransient()
    roughness = ba.Roughness(autocorr, transient)

    # Layers
    layer_vacuum = ba.Layer(vacuum)
    layer_Pd = ba.Layer(pd_mat, 12*nm, roughness)
    layer_Fe = ba.Layer(fe_mat, 100*nm, roughness)
    layer_substrate = ba.Layer(substrate_mat, roughness)

    # Multilayer
    sample = ba.Sample()
    sample.addLayer(layer_vacuum)
    sample.addLayer(layer_Pd)
    sample.addLayer(layer_Fe)
    sample.addLayer(layer_substrate)

    return sample


def simulate(p_dir, a_dir, p_eff, a_trans, a_supp, title):
    sample = get_sample()

    qzs = numpy.linspace(0.1, 1.5, 1500)
    distr = ba.DistributionGaussian(0., 1., 25, 4.)

    scan = ba.QzScan(qzs)
    scan.setAbsoluteQResolution(distr, 0.008)

    scan.setPolarization(p_dir*p_eff)
    scan.setAnalyzer(a_dir, a_trans, a_supp)

    simulation = ba.SpecularSimulation(scan, sample)
    simulation.setBackground(ba.ConstantBackground(1e-7))

    result = simulation.simulate()
    result.setTitle(title)

    return result


if __name__ == '__main__':
    ba.showSample3D(get_sample(), sample_size=80*nm, seed=0)
    # polarizer and analyzer efficiencies
    p_eff = 0.986
    a_trans = 0.970
    a_supp = 0.015

    results_pp = simulate(R3(0, +1, 0), R3(0, +1, 0), p_eff, a_trans, a_supp, "$++$")
    results_pm = simulate(R3(0, +1, 0), R3(0, -1, 0), p_eff, a_trans, a_supp, "$+-$")
    results_mp = simulate(R3(0, -1, 0), R3(0, +1, 0), p_eff, a_trans, a_supp, "$-+$")
    results_mm = simulate(R3(0, -1, 0), R3(0, -1, 0), p_eff, a_trans, a_supp, "$--$")

    results = [results_pp, results_pm, results_mp, results_mm]
    bp.plot_multicurve(results)
    bp.plt.show()



#  **************************************************************************  #
"""
#   BornAgain: simulate and fit reflection and scattering
#
#   @file      Wrap/Python/src/bornagain/ba_fit.py
#   @brief     Pure-Python FitObjective and IterationInfo, replacing Sim/Fitting C++.
#
#   @homepage  http://apps.jcns.fz-juelich.de/BornAgain
#   @license   GNU General Public License v3 or higher (see COPYING)
#   @copyright Forschungszentrum Juelich GmbH 2026
#   @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
"""
#  **************************************************************************  #

import numpy as np


class IterationInfo:
    """Stores fit iteration info to track fit flow from observers."""

    def __init__(self):
        self._count = 0
        self._chi2 = 0.0
        self._params = {}

    def update(self, params, chi2):
        self._count += 1
        self._chi2 = chi2
        self._params = dict(params)

    def iterationCount(self):
        """Returns current number of minimizer iterations."""
        return self._count

    def chi2(self):
        return self._chi2

    def parameterMap(self):
        """Returns map of fit parameter names and current values."""
        return self._params


class FitObjective:
    """
    Orchestrates fitting: runs simulations, compares to data, computes chi².

    Supports simultaneous fitting of several data sets, as demonstrated
    in example fit/scatter2d/multiple_datasets.py.
    """

    def __init__(self):
        self._pairs = []
        self._observers = []
        self._info = IterationInfo()
        self._metric = 'poisson'
        self._norm = 'l2'
        self._interrupted = False

    def addFitPair(self, sim_fn, exp_data, weight=1.0):
        """
        Adds a simulation builder and experimental data to the fit objective.

        Arguments:

        sim_fn:   a user-defined function that takes a parameter dict as input
                  and returns a BornAgain Simulation object.

        exp_data: a Datafield containing experimental data.

        weight:   user-defined weight of the dataset (default 1.0).
        """
        self._pairs.append({
            'sim_fn': sim_fn,
            'raw_data': exp_data,
            'weight': weight,
            'sim_result': None,
            'exp_result': None,
        })

    def setObjectiveMetric(self, metric, norm='l2'):
        """
        Sets the objective metric.

        metric: 'poisson' (default), 'chi2', 'log', or 'reldiff'
        norm:   'l2' (default) or 'l1'
        """
        self._metric = metric.lower()
        self._norm = norm.lower()

    def evaluate(self, params):
        """Runs simulations and returns the scalar metric value."""
        if self._interrupted:
            raise RuntimeError("Fitting was interrupted by the user.")
        if not self._pairs:
            raise RuntimeError(
                "Cannot start fit as no simulation/data pairs are defined."
            )

        params_dict = FitObjective._convert_params(params)
        self._exec_simulations(params_dict)
        chi2 = self._compute_metric()
        self._info.update(params_dict, chi2)

        count = self._info.iterationCount()
        for every_nth, callback in self._observers:
            if count % every_nth == 0:
                callback(self)

        return chi2

    def evaluate_residuals(self, params):
        """
        Runs simulations and returns the array of residuals exp - sim.

        This is the function to pass to lmfit.minimize for
        Levenberg-Marquardt fitting.
        """
        self.evaluate(params)
        exp = np.asarray(list(self.flatExpData()))
        sim = np.asarray(list(self.flatSimData()))
        return exp - sim

    def initPrint(self, every_nth):
        """Initializes printing to standard output on every_nth fit iteration."""
        from bornagain import ba_fitmonitor
        self.initPlot(every_nth, ba_fitmonitor.Printer())

    def initPlot(self, every_nth, callback):
        """Registers a callback to be called every every_nth fit iteration."""
        self._observers.append((every_nth, callback))

    def simulationResult(self, i=0):
        """Returns simulation result for the i-th dataset as a Datafield."""
        return self._pairs[i]['sim_result']

    def experimentalData(self, i=0):
        """Returns experimental data for the i-th dataset as a Datafield."""
        return self._pairs[i]['exp_result']

    def relativeDifference(self, i=0):
        """Returns relative difference between sim and exp as a Datafield."""
        sim = self._pairs[i]['sim_result']
        exp = self._pairs[i]['exp_result']
        sim_v = np.asarray(list(sim.flatVector()))
        exp_v = np.asarray(list(exp.flatVector()))
        denom = np.abs(sim_v) + np.abs(exp_v)
        rel = np.where(denom > 0, 2.0*np.abs(sim_v - exp_v)/denom, 0.0)
        import bornagain as ba
        return ba.Datafield(sim.frame(), list(rel))

    def absoluteDifference(self, i=0):
        """Returns absolute difference between sim and exp as a Datafield."""
        sim = self._pairs[i]['sim_result']
        exp = self._pairs[i]['exp_result']
        sim_v = np.asarray(list(sim.flatVector()))
        exp_v = np.asarray(list(exp.flatVector()))
        diff = np.abs(sim_v - exp_v)
        import bornagain as ba
        return ba.Datafield(sim.frame(), list(diff))

    def flatSimData(self):
        """Returns merged simulated intensities from all datasets."""
        result = []
        for pair in self._pairs:
            result.extend(pair['sim_result'].flatVector())
        return result

    def flatExpData(self):
        """Returns merged experimental data from all datasets."""
        result = []
        for pair in self._pairs:
            result.extend(pair['exp_result'].flatVector())
        return result

    def nPairs(self):
        """Returns number of simulation/data pairs."""
        return len(self._pairs)

    def iterationInfo(self):
        """Returns current IterationInfo."""
        return self._info

    def interruptFitting(self):
        """Requests that fitting be stopped after the current iteration."""
        self._interrupted = True

    def isInterrupted(self):
        """Returns True if fitting was interrupted."""
        return self._interrupted

    @staticmethod
    def _convert_params(params):
        """Converts lmfit.Parameters or plain dict to a plain Python dict."""
        if str(getattr(params, '__module__', '')) == "lmfit.parameter":
            return {p: params[p].value for p in params}
        if hasattr(params, 'items'):
            return {k: float(v) for k, v in params.items()}
        return dict(params)

    def _exec_simulations(self, params_dict):
        for pair in self._pairs:
            try:
                sim = pair['sim_fn'](params_dict)
            except Exception as exc:
                raise RuntimeError(
                    f"FitObjective: simulation builder raised "
                    f"{type(exc).__name__}: {exc}") from exc
            pair['sim_result'] = sim.simulate()
            if pair['exp_result'] is None:
                pair['exp_result'] = _reposition_exp(sim, pair['raw_data'])

    def _compute_metric(self):
        metric = self._metric
        norm = self._norm
        eps = np.finfo(float).tiny
        total = 0.0
        for pair in self._pairs:
            sim_v = np.asarray(list(pair['sim_result'].flatVector()))
            exp_v = np.asarray(list(pair['exp_result'].flatVector()))
            errs = np.asarray(list(pair['exp_result'].errorSigmas()))
            has_unc = len(errs) == len(sim_v) and len(errs) > 0

            mask = exp_v >= 0.0

            if metric == 'chi2':
                if has_unc:
                    unc_mask = mask & (errs > 0.0)
                    r = np.where(unc_mask, (exp_v - sim_v)/errs, 0.0)
                else:
                    r = np.where(mask, exp_v - sim_v, 0.0)
            elif metric in ('poisson', 'poisson-like'):
                if has_unc:
                    unc_mask = mask & (errs > 0.0)
                    r = np.where(unc_mask, (exp_v - sim_v)/errs, 0.0)
                else:
                    variance = np.maximum(1.0, sim_v)
                    r = np.where(mask, (sim_v - exp_v)/np.sqrt(variance),
                                 0.0)
            elif metric == 'log':
                if has_unc:
                    ln10 = np.log(10.0)
                    unc_mask = mask & (errs > 0.0)
                    sim_s = np.maximum(eps, sim_v)
                    exp_s = np.maximum(eps, exp_v)
                    safe_errs = np.where(unc_mask, errs, 1.0)
                    val = (np.log10(sim_s) -
                           np.log10(exp_s))*exp_s*ln10/safe_errs
                    r = np.where(unc_mask, val, 0.0)
                else:
                    sim_s = np.maximum(eps, sim_v)
                    exp_s = np.maximum(eps, exp_v)
                    r = np.where(mask,
                                 np.log10(sim_s) - np.log10(exp_s), 0.0)
            elif metric in ('reldiff', 'relative'):
                if has_unc:
                    unc_mask = mask & (errs > 0.0)
                    r = np.where(unc_mask, (exp_v - sim_v)/errs, 0.0)
                else:
                    sim_s = np.maximum(eps, sim_v)
                    exp_s = np.maximum(eps, exp_v)
                    r = np.where(mask, (exp_s - sim_s)/(exp_s + sim_s),
                                 0.0)
            else:
                raise ValueError(
                    f"Unknown metric: {metric!r}. "
                    "Valid options: 'chi2', 'poisson', 'log', 'reldiff'.")

            if norm == 'l2':
                contrib = float(np.sum(r*r))
            elif norm == 'l1':
                contrib = float(np.sum(np.abs(r)))
            else:
                raise ValueError(
                    f"Unknown norm: {norm!r}. Valid options: 'l2', 'l1'.")

            total += pair['weight']*contrib

        return total


def _reposition_exp(simulation, data):
    """
    Repositions experimental data to match the simulation's detector ROI frame.

    For ScatteringSimulation, crops or maps the data to the active ROI pixels.
    For specular and other simulations, returns the data unchanged.
    """
    if not hasattr(simulation, 'detector'):
        return data

    det = simulation.detector()
    clipped_frame = det.clippedFrame()
    active_indices = list(det.activeIndices())

    n = clipped_frame.size()
    flat_data = list(data.flatVector())
    flat_errs = list(data.errorSigmas())
    has_errors = len(flat_errs) == len(flat_data) and len(flat_errs) > 0

    values = [0.0]*n
    errors = [0.0]*n if has_errors else []

    if data.frame().hasSameSizes(clipped_frame):
        # Data is already ROI-sized
        for i in active_indices:
            values[i] = flat_data[i]
            if has_errors:
                errors[i] = flat_errs[i]
    else:
        # Expect data to have full detector shape
        data_matches_det = (data.rank() == 2
                            and data.axis(0).size() == det.axis(0).size()
                            and data.axis(1).size() == det.axis(1).size())
        if not data_matches_det:
            raise RuntimeError(
                "FitObjective: Detector and experimental data have different shape"
            )
        for i in active_indices:
            full_i = det.roiToFullIndex(i)
            values[i] = flat_data[full_i]
            if has_errors:
                errors[i] = flat_errs[full_i]

    import bornagain as ba
    if has_errors:
        return ba.Datafield(clipped_frame, values, errors)
    return ba.Datafield(clipped_frame, values)



#!/usr/bin/env python3
"""
Dilute cylinders on substrate.
Test to investigate intensity changes with different numbers of slices.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm
import matplotlib.pyplot as plt


def get_sample(n_slices=1):
    # Materials (with absorption for sharper tests)
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 6e-04, 2e-05)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-06, 2e-08)
    vacuum = ba.Vacuum()

    # Particle
    ff = ba.Cylinder(5*nm, 5*nm)
    particle = ba.Particle(particle_mat, ff)

    # Layers
    layer_1 = ba.Layer(vacuum)
    if n_slices > 1:
        layer_1.setNumberOfSlices(n_slices)
    layer_1.deposit2D(ba.Dilute2D(0.001, particle))
    layer_2 = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def simulate_horizontal_cut(n_slices=1):
    """Simulate with Nx1 detector for horizontal cut at alpha_f=0.92 deg."""
    sample = get_sample(n_slices)
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 11
    # Single bin centered at alpha_f=0.92 deg
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, 1, 0.91*deg, 0.93*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation.simulate()


def simulate_vertical_cut(n_slices=1):
    """Simulate with 1xN detector for vertical cut at phi_f=0 deg."""
    sample = get_sample(n_slices)
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 11
    # Single bin centered at phi_f=0 deg
    detector = ba.SphericalDetector(1, -0.01*deg, 0.01*deg, n, 0, 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation.simulate()


# Color cycle for different N values
colors = ['C0', 'C1', 'C2', 'C3']


if __name__ == '__main__':
    # Slice counts to compare
    slice_counts = [1, 2, 5, 20]

    # Create figure with 1 row, 2 columns
    fig, axes = plt.subplots(1, 2, figsize=(12, 5))

    # Horizontal cuts at alpha_f=0.92 deg
    ax = axes[0]
    for i, n in enumerate(slice_counts):
        result = simulate_horizontal_cut(n_slices=n)
        ax.semilogy(result.xCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title('Horizontal cut at alpha_f=0.92 deg')
    ax.set_xlabel('phi_f (deg)')
    ax.set_ylabel('Intensity')
    ax.legend()

    # Vertical cuts at phi_f=0 deg
    ax = axes[1]
    for i, n in enumerate(slice_counts):
        result = simulate_vertical_cut(n_slices=n)
        ax.semilogy(result.yCenters(), result.flatVector(),
                    label=f'N={n}', color=colors[i])
    ax.set_title('Vertical cut at phi_f=0 deg')
    ax.set_xlabel('alpha_f (deg)')
    ax.set_ylabel('Intensity')
    ax.legend()

    plt.tight_layout()
    # Persistence test for N=20 slices
    from bornagain import ba_check
    ba_check.persistence_test(simulate_horizontal_cut(n_slices=20))



#!/usr/bin/env python3
"""
Dense film of small spheres: scattering intensity for different film thicknesses.
Uses the Percus-Yevick approximation.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, nm3


def get_sample(thickness):
    # Materials
    particle_color = (0.86, 0.24, 0.18)
    particle_mat = ba.RefractiveMaterial("Particle", particle_color, 1e-5, 0)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 1e-6, 0)
    vacuum = ba.Vacuum()

    # Particles
    ff = ba.Sphere(4*nm)
    particle = ba.Particle(particle_mat, ff)

    # Layers
    layer_1 = ba.Layer(vacuum)
    layer_2 = ba.Layer(vacuum, thickness)
    layer_3 = ba.Layer(substrate_mat)
    layer_2.fill3D(ba.Dilute3D(1e-8, particle))

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)
    sample.addLayer(layer_3)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 11
    # Just compute a 1d cut at phi=0
    detector = ba.SphericalDetector(1, -1*deg, 1*deg, n, 0., 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    thicknesses = [10, 31.6, 100, 316, 1000]
    samples = [get_sample(t) for t in thicknesses]
    results = [get_simulation(sample).simulate() for sample in samples]
    from bornagain import ba_check
    ba_check.persistence_test(results)



#!/usr/bin/env python3
"""
Cylindrical particle made from two materials.
Particle crosses air/substrate interface.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, nm2, R3


def get_sample():
    """
    A multilayer with substrate/air layers.
    Vacuum layer contains cylindrical particles made of two materials.
    Particle shifted down to cross interface.
    """

    # Materials
    ag_color = (0.86, 0.24, 0.18)
    ag_mat = ba.RefractiveMaterial("Ag", ag_color, 1.245e-05, 5.419e-07)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 3.212e-06,
                                               3.244e-08)
    teflon_color = (0.93, 0.72, 0.25)
    teflon_mat = ba.RefractiveMaterial("Teflon", teflon_color, 2.9e-06, 6.019e-09)
    vacuum = ba.Vacuum()

    # Form factors
    ff_1 = ba.Cylinder(10*nm, 4*nm)
    ff_2 = ba.Cylinder(10*nm, 10*nm)

    # Particles
    subparticle_1 = ba.Particle(ag_mat, ff_1)
    subparticle_1.translate(R3(0, 0, 10*nm))
    subparticle_2 = ba.Particle(teflon_mat, ff_2)

    # Composition of particles at specific positions
    particle = ba.Compound()
    particle.addComponent(subparticle_1)
    particle.addComponent(subparticle_2)
    particle.translate(R3(0, 0, -10*nm))

    # Layers
    layer_1 = ba.Layer(vacuum)
    layer_1.deposit2D(ba.Dilute2D(0.0001/nm2, particle))
    layer_2 = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 100
    detector = ba.SphericalDetector(n, -1*deg, 1*deg, n, 0., 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    plotargs = bp.parse_commandline()
    bp.plot_datafield(result, unit_aspect=1, **plotargs)
    bp.export(**plotargs)



#!/usr/bin/env python3
"""
Basic example how to fit specular data.
The sample consists of twenty alternating Ti and Ni layers.
Reference data was generated with GenX.
We fit just one parameter, the thickness of the Ti layers,
which has an original value of 30 angstroms.
"""

import bornagain as ba, numpy as np, os
from bornagain import angstrom, ba_fitmonitor, nm
import lmfit


def load_data():
    datadir = os.getenv('BA_DATA_DIR', '')
    if not datadir:
        raise Exception("Environment variable BA_DATA_DIR not set")
    fname = os.path.join(datadir, "specular/genx_alternating_layers.dat.gz")

    flags = ba.ImportSettings1D("2alpha (deg)", "#", "", 1, 2)
    return ba.readData1D(fname, ba.csv1D, flags)


def get_sample(P):
    # Materials
    vacuum = ba.Vacuum()
    ti_color = (0.05, 0.62, 0.55)
    ti_mat = ba.SLDMaterial("Ti", ti_color, -1.9493e-06, 0)
    ni_color = (0.93, 0.48, 0.14)
    ni_mat = ba.SLDMaterial("Ni", ni_color, 9.4245e-06, 0)
    si_color = (0.30, 0.62, 0.86)
    si_mat = ba.SLDMaterial("Si", si_color, 2.0704e-06, 0)

    # Layers
    layer_Ti = ba.Layer(ti_mat, P["thickness_Ti"])
    layer_Ni = ba.Layer(ni_mat, 70*angstrom)

    # Periodic stack
    n_repetitions = 10
    stack = ba.LayerStack(n_repetitions)
    stack.addLayer(layer_Ti)
    stack.addLayer(layer_Ni)

    # Sample
    sample = ba.Sample()
    sample.addLayer(ba.Layer(vacuum))
    sample.addStack(stack)
    sample.addLayer(ba.Layer(si_mat))

    return sample


def get_simulation(P):
    scan = ba.AlphaScan(data.xAxis())
    scan.setWavelength(1.54*angstrom)
    sample = get_sample(P)

    return ba.SpecularSimulation(scan, sample)


if __name__ == '__main__':
    data = load_data()

    fit_objective = ba.FitObjective()
    fit_objective.addFitPair(get_simulation, data, 1)

    fit_objective.initPrint(10)
    plot_observer = ba_fitmonitor.PlotterSpecular(pause=0.5)
    fit_objective.initPlot(10, plot_observer)

    P = lmfit.Parameters()
    P.add("thickness_Ti", 50*angstrom, min=10*angstrom, max=60*angstrom)

    result = lmfit.minimize(fit_objective.evaluate_residuals, P)
    print(lmfit.fit_report(result))

    plot_observer.show()



#!/usr/bin/env python3
"""
Reflectivity with and without polarization analyzer.
"""
import unittest
import bornagain as ba
from bornagain import deg, nm, R3

from math import sin, cos


def get_sample():
    # Materials
    Bmag = 1e8
    Bangle = 60*deg
    B = R3(Bmag*cos(Bangle), Bmag*sin(Bangle), 0)
    layer_color = (0.05, 0.62, 0.55)
    layer_mat = ba.SLDMaterial("Layer", layer_color, 0.0001, 1e-08, B)
    substrate_color = (0.55, 0.55, 0.55)
    substrate_mat = ba.SLDMaterial("Substrate", substrate_color, 7e-05, 2e-06)
    vacuum = ba.Vacuum()

    # Layers
    layer_1 = ba.Layer(vacuum)
    layer_2 = ba.Layer(layer_mat, 10*nm)
    layer_3 = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)
    sample.addLayer(layer_3)

    return sample


def run_simulation(polarizer_vec, analyzer_vec=None):
    sample = get_sample()

    n = 50
    scan = ba.AlphaScan(n, 5*deg/n, 5*deg)
    scan.setWavelength(0.15*nm)

    # adding polarizer and analyzer operator
    scan.setPolarization(polarizer_vec)
    if analyzer_vec:
        scan.setAnalyzer(analyzer_vec)

    simulation = ba.SpecularSimulation(scan, sample)

    result = simulation.simulate()

    return result.xCenters(), result.intensities()


class PolarizedNoAnalyzerTest(unittest.TestCase):

    def testPolarizedNoAnalyzer(self):
        q, results_pp = run_simulation(R3(0, +1, 0), R3(0, +1, 0))
        q, results_mm = run_simulation(R3(0, -1, 0), R3(0, -1, 0))

        q, results_pm = run_simulation(R3(0, +1, 0), R3(0, -1, 0))
        q, results_mp = run_simulation(R3(0, -1, 0), R3(0, +1, 0))

        r_plus = results_pp + results_pm
        r_minus = results_mm + results_mp

        # same result, but need half the computational time
        q, results_p = run_simulation(R3(0, +1, 0))
        q, results_m = run_simulation(R3(0, -1, 0))

        self.assertTrue(ba.checkRelVecDifference(results_pp + results_pm, results_p, 1e-14))
        self.assertTrue(ba.checkRelVecDifference(results_mp + results_mm, results_m, 1e-14))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
Cylindrical mesocrystal with hexagonal lattice of soft spheres on a substrate.
Basis made of two different materials.
With slicing.
"""
import math
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, nm2, R3


def get_sample():
    # Materials
    particle1_color = (0.86, 0.24, 0.18)
    particle1_mat = ba.RefractiveMaterial("Particle1", particle1_color, 4e-05, 3e-08)
    particle2_color = (0.25, 0.65, 0.35)
    particle2_mat = ba.RefractiveMaterial("Particle2", particle2_color, 8e-06, 5e-08)
    substrate_color = (0.28, 0.57, 0.82)
    substrate_mat = ba.RefractiveMaterial("Substrate", substrate_color, 6e-06, 2e-08)
    vacuum = ba.Vacuum()

    # Basis particles
    inner_ff = ba.FuzzySphere(4.7*nm, .4*nm)

    inner_particle1 = ba.Particle(particle1_mat, inner_ff)
    inner_particle2 = ba.Particle(particle2_mat, inner_ff)

    # 3D lattice
    lattice_a = 6.21
    lattice_c = 6.57
    sigma_a = 1.16
    position_variance = sigma_a/math.sqrt(3)

    lattice = ba.HexagonalLattice(lattice_a*2, lattice_c*2*2.3)

    bas_a = lattice.basisVectorA()
    bas_b = lattice.basisVectorB()
    bas_c = lattice.basisVectorC()

    # Basis composition
    position_0 = R3(0, 0, 0)
    position_1 = 1.0/3*(2*bas_a + bas_b + bas_c)
    basis = ba.Compound()
    basis.addComponent(inner_particle1, position_0)
    basis.addComponent(inner_particle2, position_1)

    # Crystal
    crystal = ba.Crystal(basis, lattice, position_variance)

    # Mesocrystal: shaped crystal
    outer_ff = ba.Cylinder(20*nm, 50*nm)
    outer_particle = ba.Mesocrystal(crystal, outer_ff)

    # Layers
    layer_1 = ba.Layer(vacuum)
    layer_1.deposit2D(ba.Dilute2D(0.0001/nm2, outer_particle))
    layer_1.setNumberOfSlices(10)
    layer_2 = ba.Layer(substrate_mat)

    # Sample
    sample = ba.Sample()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1*nm, 0.2*deg)
    n = 200
    detector = ba.SphericalDetector(n, -2*deg, 2*deg, n, 0, 2*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    sample = get_sample()
    ba.showSample3D(sample, sample_size=250*nm, seed=0)
    simulation = get_simulation(sample)
    result = simulation.simulate()
    bp.plot_datafield(result, unit_aspect=1)
    bp.plt.show()



#!/usr/bin/env python3
"""
Basic example of specular reflectometry simulation,
using a standard sample from Wrap/Python/std_samples.py.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, angstrom, std_samples, nm


def get_sample():
    return std_samples.alternating_layers()


def get_simulation(sample):
    n = 50
    scan = ba.AlphaScan(n, 2*deg/n, 2*deg)
    scan.setWavelength(1.54*angstrom)
    return ba.SpecularSimulation(scan, sample)


if __name__ == '__main__':
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    from bornagain import ba_check
    ba_check.persistence_test(result)



#!/usr/bin/env python3
"""
Same as CylindersAndSpecular.py, but with beam and detector
both rotated by phi = 1.34 deg.

This example tests that indexOfSpecular correctly handles non-zero phi.
The simulation result should be identical to CylindersAndSpecular.py.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm

phi = 1.34*deg


def get_sample():
    """
    Standard sample model: dilute random assembly of cylinders on a substrate.
    """
    from bornagain import std_samples
    return std_samples.cylinders()


def get_simulation(sample):
    # Beam with non-zero phi
    wavelength = 0.1*nm
    alpha_i = 0.2*deg
    beam = ba.Beam(1e9, wavelength, alpha_i, phi)

    # Detector rotated by phi, otherwise same as in CylindersAndSpecular
    n = <%= test_mode ? 11 : 200 %>
    detector = ba.SphericalDetector(n, -2*deg + phi, 2*deg + phi, n, 0., 3*deg)

    simulation = ba.ScatteringSimulation(beam, sample, detector)
    simulation.options().setIncludeSpecular(True)
    return simulation


<%= include_partial("template/main/runsim_unitaspect.py",
    "ba.showSample3D(sample, sample_size=120*nm, seed=0)") -%>
